"""
https://github.github.com/gfm/#links
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_reference_links_535():
    """
    Test case 535:  Here is a simple example:
    """

    # Arrange
    source_markdown = """[foo][bar]

[bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/url:title:::bar:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_535a():
    """
    Test case 535:  Here is a simple example:
    """

    # Arrange
    source_markdown = """[foo][bar]\a

[bar]: /url "title"
""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):: ]",
        "[link(1,1):full:/url:title:::bar:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_535b():
    """
    Test case 535:  Here is a simple example:
    """

    # Arrange
    source_markdown = """[foo][bar] abc

[bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/url:title:::bar:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[text(1,11): abc:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a> abc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_536():
    """
    Test case 536:  (part 1) The link text may contain balanced brackets, but not unbalanced ones, unless they are escaped:
    """

    # Arrange
    source_markdown = """[link [foo [bar]]][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/uri::::ref:link [foo [bar]]:::::]",
        "[text(1,2):link :]",
        "[text(1,7):[:]",
        "[text(1,8):foo :]",
        "[text(1,12):[:]",
        "[text(1,13):bar:]",
        "[text(1,16):]:]",
        "[text(1,17):]:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri">link [foo [bar]]</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_537():
    """
    Test case 537:  (part 2) The link text may contain balanced brackets, but not unbalanced ones, unless they are escaped:
    """

    # Arrange
    source_markdown = """[link \\[bar][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/uri::::ref:link \\[bar:::::]",
        "[text(1,2):link \\\b[bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri">link [bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_538():
    """
    Test case 538:  (part 1) The link text may contain inline content:
    """

    # Arrange
    source_markdown = """[link *foo **bar** `#`*][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/uri::::ref:link *foo **bar** `#`*:::::]",
        "[text(1,2):link :]",
        "[emphasis(1,7):1:*]",
        "[text(1,8):foo :]",
        "[emphasis(1,12):2:*]",
        "[text(1,14):bar:]",
        "[end-emphasis(1,17)::]",
        "[text(1,19): :]",
        "[icode-span(1,20):#:`::]",
        "[end-emphasis(1,23)::]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri">link <em>foo <strong>bar</strong> <code>#</code></em></a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_539():
    """
    Test case 539:  (part 2) The link text may contain inline content:
    """

    # Arrange
    source_markdown = """[![moon](moon.jpg)][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/uri::::ref:![moon](moon.jpg):::::]",
        "[image(1,2):inline:moon.jpg::moon::::moon:False::::]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri"><img src="moon.jpg" alt="moon" /></a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_540():
    """
    Test case 540:  (part 1) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """[foo [bar](/uri)][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        "[link(1,6):inline:/uri:::::bar:False::::]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[text(1,17):]:]",
        "[link(1,18):shortcut:/uri:::::ref:::::]",
        "[text(1,19):ref:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p>[foo <a href="/uri">bar</a>]<a href="/uri">ref</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_541():
    """
    Test case 541:  (part 2) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """[foo *bar [baz][ref]*][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):1:*]",
        "[text(1,7):bar :]",
        "[link(1,11):full:/uri::::ref:baz:::::]",
        "[text(1,12):baz:]",
        "[end-link::]",
        "[end-emphasis(1,21)::]",
        "[text(1,22):]:]",
        "[link(1,23):shortcut:/uri:::::ref:::::]",
        "[text(1,24):ref:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = (
        """<p>[foo <em>bar <a href="/uri">baz</a></em>]<a href="/uri">ref</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_542():
    """
    Test case 542:  (part 1) The following cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """*[foo*][ref]

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[link(1,2):full:/uri::::ref:foo*:::::]",
        "[text(1,3):foo:]",
        "[text(1,6):*:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p>*<a href="/uri">foo*</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_543():
    """
    Test case 543:  (part 2) The following cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """[foo *bar][ref]*

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/uri::::ref:foo *bar:::::]",
        "[text(1,2):foo :]",
        "[text(1,6):*:]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[text(1,16):*:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri">foo *bar</a>*</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_544():
    """
    Test case 544:  (part 1) The following cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """[foo <bar attr="][ref]">

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        '[raw-html(1,6):bar attr="][ref]"]',
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p>[foo <bar attr="][ref]"></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_545():
    """
    Test case 545:  (part 2) The following cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """[foo`][ref]`

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[icode-span(1,5):][ref]:`::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p>[foo<code>][ref]</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_546():
    """
    Test case 546:  (part 3) The following cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """[foo<http://example.com/?search=][ref]>

[ref]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[uri-autolink(1,5):http://example.com/?search=][ref]]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref:: :/uri:::::]",
    ]
    expected_gfm = """<p>[foo<a href="http://example.com/?search=%5D%5Bref%5D">http://example.com/?search=][ref]</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_547():
    """
    Test case 547:  Matching is case-insensitive:
    """

    # Arrange
    source_markdown = """[foo][BaR]

[bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/url:title:::BaR:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_548():
    """
    Test case 548:  Unicode case fold is used:
    """

    # Arrange
    source_markdown = """[ẞ]

[SS]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:::::ẞ:::::]",
        "[text(1,2):ẞ:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ss:SS: :/url:::::]",
    ]
    expected_gfm = """<p><a href="/url">ẞ</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_549():
    """
    Test case 549:  Consecutive internal whitespace is treated as one space for purposes of determining matching:
    """

    # Arrange
    source_markdown = """[Foo
  bar]: /url

[Baz][Foo bar]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo bar:Foo\n  bar: :/url:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):]",
        "[link(4,1):full:/url::::Foo bar:Baz:::::]",
        "[text(4,2):Baz:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url">Baz</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_550():
    """
    Test case 550:  (part 1) No whitespace is allowed between the link text and the link label:
    """

    # Arrange
    source_markdown = """[foo] [bar]

[bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6): :]",
        "[link(1,7):shortcut:/url:title::::bar:::::]",
        "[text(1,8):bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p>[foo] <a href="/url" title="title">bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_551():
    """
    Test case 551:  (part 2) No whitespace is allowed between the link text and the link label:
    """

    # Arrange
    source_markdown = """[foo]
[bar]

[bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):\n::\n]",
        "[link(2,1):shortcut:/url:title::::bar:::::]",
        "[text(2,2):bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        '[link-ref-def(4,1):True::bar:: :/url:: :title:"title":]',
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<p>[foo]
<a href="/url" title="title">bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_552():
    """
    Test case 552:  When there are multiple matching link reference definitions, the first is used:
    """

    # Arrange
    source_markdown = """[foo]: /url1

[foo]: /url2

[bar][foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url1:::::]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):False::foo:: :/url2:::::]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[link(5,1):full:/url1::::foo:bar:::::]",
        "[text(5,2):bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url1">bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_553():
    """
    Test case 553:  Note that matching is performed on normalized strings, not parsed inline content. So the following does not match, even though the labels define equivalent inline content:
    """

    # Arrange
    source_markdown = """[bar][foo\\!]

[foo!]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):bar:]",
        "[text(1,5):]:]",
        "[text(1,6):[:]",
        "[text(1,7):foo\\\b!:]",
        "[text(1,12):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo!:: :/url:::::]",
    ]
    expected_gfm = """<p>[bar][foo!]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_554():
    """
    Test case 554:  (part 1) Link labels cannot contain brackets, unless they are backslash-escaped:
    """

    # Arrange
    source_markdown = """[foo][ref[]

[ref[]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):[:]",
        "[text(1,7):ref:]",
        "[text(1,10):[:]",
        "[text(1,11):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):ref:]",
        "[text(3,5):[:]",
        "[text(3,6):]:]",
        "[text(3,7):: /uri:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo][ref[]</p>
<p>[ref[]: /uri</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_555():
    """
    Test case 555:  (part 2) Link labels cannot contain brackets, unless they are backslash-escaped:
    """

    # Arrange
    source_markdown = """[foo][ref[bar]]

[ref[bar]]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):[:]",
        "[text(1,7):ref:]",
        "[text(1,10):[:]",
        "[text(1,11):bar:]",
        "[text(1,14):]:]",
        "[text(1,15):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):ref:]",
        "[text(3,5):[:]",
        "[text(3,6):bar:]",
        "[text(3,9):]:]",
        "[text(3,10):]:]",
        "[text(3,11):: /uri:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo][ref[bar]]</p>
<p>[ref[bar]]: /uri</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_556():
    """
    Test case 556:  (part 3) Link labels cannot contain brackets, unless they are backslash-escaped:
    """

    # Arrange
    source_markdown = """[[[foo]]]

[[[foo]]]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):[:]",
        "[text(1,3):[:]",
        "[text(1,4):foo:]",
        "[text(1,7):]:]",
        "[text(1,8):]:]",
        "[text(1,9):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):[:]",
        "[text(3,3):[:]",
        "[text(3,4):foo:]",
        "[text(3,7):]:]",
        "[text(3,8):]:]",
        "[text(3,9):]:]",
        "[text(3,10):: /url:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[[[foo]]]</p>
<p>[[[foo]]]: /url</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_557():
    """
    Test case 557:  (part 4) Link labels cannot contain brackets, unless they are backslash-escaped:
    """

    # Arrange
    source_markdown = """[foo][ref\\[]

[ref\\[]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/uri::::ref\\[:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::ref\\[:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_558():
    """
    Test case 558:  Note that in this example ] is not backslash-escaped:
    """

    # Arrange
    source_markdown = """[bar\\\\]: /uri

[bar\\\\]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar\\\\:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar\\\\:::::]",
        "[text(3,2):bar\\\b\\:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar\\</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_558a():
    """
    Test case 558a:  variation
    """

    # Arrange
    source_markdown = """[bar&#x5C;]: /uri

[bar&#x5C;]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&#x5c;:bar&#x5C;: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar&#x5C;:::::]",
        "[text(3,2):bar\a&#x5C;\a\\\a:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar\\</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_558b():
    """
    Test case 558a:  variation
    """

    # Arrange
    source_markdown = """[bar&beta;]: /uri

[bar&beta;]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&beta;:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar&beta;:::::]",
        "[text(3,2):bar\a&beta;\aβ\a:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">barβ</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_559():
    """
    Test case 559:  (part 1) A link label must contain at least one non-whitespace character:
    """

    # Arrange
    source_markdown = """[]

[]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):]:]",
        "[text(3,3):: /uri:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[]</p>
<p>[]: /uri</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_560():
    """
    Test case 560:  (part 2) A link label must contain at least one non-whitespace character:
    """

    # Arrange
    source_markdown = """[
 ]

[
 ]: /uri"""
    expected_tokens = [
        "[para(1,1):\n ]",
        "[text(1,1):[:]",
        "[text(1,2):\n::\n]",
        "[text(2,2):]:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n ]",
        "[text(4,1):[:]",
        "[text(4,2):\n::\n]",
        "[text(5,2):]:]",
        "[text(5,3):: /uri:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[
]</p>
<p>[
]: /uri</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_561():
    """
    Test case 561:  (part 1) Thus, [foo][] is equivalent to [foo][foo].
    """

    # Arrange
    source_markdown = """[foo][]

[foo]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):collapsed:/url:title::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_562():
    """
    Test case 562:  (part 2) Thus, [foo][] is equivalent to [foo][foo].
    """

    # Arrange
    source_markdown = """[*foo* bar][]

[*foo* bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):collapsed:/url:title::::*foo* bar:::::]",
        "[emphasis(1,2):1:*]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[text(1,7): bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::*foo* bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title"><em>foo</em> bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_563():
    """
    Test case 563:  The link labels are case-insensitive:
    """

    # Arrange
    source_markdown = """[Foo][]

[foo]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):collapsed:/url:title::::Foo:::::]",
        "[text(1,2):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">Foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_564():
    """
    Test case 564:  As with full reference links, whitespace is not allowed between the two sets of brackets:
    """

    # Arrange
    source_markdown = """[foo]\a
[]

[foo]: /url "title"
""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n]",
        "[link(1,1):shortcut:/url:title::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[text(1,6):\n:: \n]",
        "[text(2,1):[:]",
        "[text(2,2):]:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        '[link-ref-def(4,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a>
[]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_565():
    """
    Test case 565:  (part 1) A shortcut reference link consists of a link label that matches a link reference definition elsewhere in the document and is not followed by [] or a link label.
    """

    # Arrange
    source_markdown = """[foo]

[foo]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:title::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_566():
    """
    Test case 566:  (part 2) A shortcut reference link consists of a link label that matches a link reference definition elsewhere in the document and is not followed by [] or a link label.
    """

    # Arrange
    source_markdown = """[*foo* bar]

[*foo* bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:title::::*foo* bar:::::]",
        "[emphasis(1,2):1:*]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[text(1,7): bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::*foo* bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title"><em>foo</em> bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_567():
    """
    Test case 567:  (part 3) A shortcut reference link consists of a link label that matches a link reference definition elsewhere in the document and is not followed by [] or a link label.
    """

    # Arrange
    source_markdown = """[[*foo* bar]]

[*foo* bar]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[link(1,2):shortcut:/url:title::::*foo* bar:::::]",
        "[emphasis(1,3):1:*]",
        "[text(1,4):foo:]",
        "[end-emphasis(1,7)::]",
        "[text(1,8): bar:]",
        "[end-link::]",
        "[text(1,13):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::*foo* bar:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p>[<a href="/url" title="title"><em>foo</em> bar</a>]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_568():
    """
    Test case 568:  (part 4) A shortcut reference link consists of a link label that matches a link reference definition elsewhere in the document and is not followed by [] or a link label.
    """

    # Arrange
    source_markdown = """[[bar [foo]

[foo]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):[:]",
        "[text(1,3):bar :]",
        "[link(1,7):shortcut:/url:::::foo:::::]",
        "[text(1,8):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :/url:::::]",
    ]
    expected_gfm = """<p>[[bar <a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_569():
    """
    Test case 569:  The link labels are case-insensitive:
    """

    # Arrange
    source_markdown = """[Foo]

[foo]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:title::::Foo:::::]",
        "[text(1,2):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/url" title="title">Foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_570():
    """
    Test case 570:  A space after the link text should be preserved:
    """

    # Arrange
    source_markdown = """[foo] bar

[foo]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[text(1,6): bar:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :/url:::::]",
    ]
    expected_gfm = """<p><a href="/url">foo</a> bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_570a():
    """
    Test case 570a:  Variation of 570 to show how link inside of link doesn't work.
    """

    # Arrange
    source_markdown = """[foo[foo]]

[foo]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[link(1,5):shortcut:/url:title::::foo:::::]",
        "[text(1,6):foo:]",
        "[end-link::]",
        "[text(1,10):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p>[foo<a href=\"/url\" title=\"title\">foo</a>]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_571():
    """
    Test case 571:  If you just want bracketed text, you can backslash-escape the opening bracket to avoid links
    """

    # Arrange
    source_markdown = """\\[foo]

[foo]: /url "title"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\\\b[foo:]",
        "[text(1,6):]:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_572():
    """
    Test case 572:  Note that this is a link, because a link label ends with the first following closing bracket:
    """

    # Arrange
    source_markdown = """[foo*]: /url

*[foo*]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo*:: :/url:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):*:]",
        "[link(3,2):shortcut:/url:::::foo*:::::]",
        "[text(3,3):foo:]",
        "[text(3,6):*:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*<a href="/url">foo*</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_573():
    """
    Test case 573:  (part 1) Full and compact references take precedence over shortcut references:
    """

    # Arrange
    source_markdown = """[foo][bar]

[foo]: /url1
[bar]: /url2"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/url2::::bar:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :/url1:::::]",
        "[link-ref-def(4,1):True::bar:: :/url2:::::]",
    ]
    expected_gfm = """<p><a href="/url2">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_574():
    """
    Test case 574:  (part 2) Full and compact references take precedence over shortcut references:
    """

    # Arrange
    source_markdown = """[foo][]

[foo]: /url1"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):collapsed:/url1:::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :/url1:::::]",
    ]
    expected_gfm = """<p><a href="/url1">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_575():
    """
    Test case 575:  (part 1) Inline links also take precedence:
    """

    # Arrange
    source_markdown = """[foo]()

[foo]: /url1"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline::::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :/url1:::::]",
    ]
    expected_gfm = """<p><a href="">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_576():
    """
    Test case 576:  (part 2) Inline links also take precedence:
    """

    # Arrange
    source_markdown = """[foo](not a link)

[foo]: /url1"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url1:::::foo:False::: :]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[text(1,6):(not a link):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :/url1:::::]",
    ]
    expected_gfm = """<p><a href="/url1">foo</a>(not a link)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_577():
    """
    Test case 577:  In the following case [bar][baz] is parsed as a reference, [foo] as normal text:
    """

    # Arrange
    source_markdown = """[foo][bar][baz]

[baz]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[link(1,6):full:/url::::baz:bar:::::]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::baz:: :/url:::::]",
    ]
    expected_gfm = """<p>[foo]<a href="/url">bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_578():
    """
    Test case 578:  Here, though, [foo][bar] is parsed as a reference, since [bar] is defined:
    """

    # Arrange
    source_markdown = """[foo][bar][baz]

[baz]: /url1
[bar]: /url2"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):full:/url2::::bar:foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[link(1,11):shortcut:/url1:::::baz:::::]",
        "[text(1,12):baz:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::baz:: :/url1:::::]",
        "[link-ref-def(4,1):True::bar:: :/url2:::::]",
    ]
    expected_gfm = """<p><a href="/url2">foo</a><a href="/url1">baz</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_579():
    """
    Test case 579:  Here [foo] is not parsed as a shortcut reference, because it is followed by a link label (even though [bar] is not defined):
    """

    # Arrange
    source_markdown = """[foo][bar][baz]

[baz]: /url1
[foo]: /url2"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[link(1,6):full:/url1::::baz:bar:::::]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::baz:: :/url1:::::]",
        "[link-ref-def(4,1):True::foo:: :/url2:::::]",
    ]
    expected_gfm = """<p>[foo]<a href="/url1">bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_01():
    """
    Test case extra 01:  variation on 644
    """

    # Arrange
    source_markdown = """[foo <!-- this is a
comment - with hyphen --> bar]: /uri

[foo <!-- this is a
comment - with hyphen --> bar]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo <!-- this is a comment - with hyphen --> bar:foo <!-- this is a\ncomment - with hyphen --> bar: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[link(4,1):shortcut:/uri:::::foo <!-- this is a\ncomment - with hyphen --> bar:::::]",
        "[text(4,2):foo :]",
        "[raw-html(4,6):!-- this is a\ncomment - with hyphen --]",
        "[text(5,26): bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">foo <!-- this is a
comment - with hyphen --> bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_02():
    """
    Test case extra 02:  variation on 345
    """

    # Arrange
    source_markdown = """[foo ``
foo
bar\a\a
baz
`` bar]: /uri

[foo ``
foo
bar\a\a
baz
`` bar]""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[link-ref-def(1,1):True::foo `` foo bar baz `` bar:foo ``\nfoo\nbar  \nbaz\n`` bar: :/uri:::::]",
        "[BLANK(6,1):]",
        "[para(7,1):\n\n\n\n]",
        "[link(7,1):shortcut:/uri:::::foo ``\nfoo\nbar  \nbaz\n`` bar:::::]",
        "[text(7,2):foo :]",
        "[icode-span(7,6):foo\a\n\a \abar  \a\n\a \abaz:``:\a\n\a \a:\a\n\a \a]",
        "[text(11,3): bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">foo <code>foo bar   baz</code> bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03x():
    """
    Test case extra 03x:  variation on 558
    """

    # Arrange
    source_markdown = """[bar\\foo]: /uri

[bar\\foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar\\foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar\\foo:::::]",
        "[text(3,2):bar\\foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar\\foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03xa():
    """
    Test case extra 03xa:  variation
    """

    # Arrange
    source_markdown = """[xx[bar\\foo]yy](/uri)

[bar\\foo]: /uri1"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        "[link(1,4):shortcut:/uri1:::::bar\\foo:::::]",
        "[text(1,5):bar\\foo:]",
        "[end-link::]",
        "[text(1,13):yy:]",
        "[text(1,15):]:]",
        "[text(1,16):(/uri):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar\\foo:: :/uri1:::::]",
    ]
    expected_gfm = """<p>[xx<a href="/uri1">bar\\foo</a>yy](/uri)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03xb():
    """
    Test case extra 03xb:  variation
    """

    # Arrange
    source_markdown = """![xx[bar\\foo]yy](/uri)

[bar\\foo]: /uri1"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:/uri::xxbar\\fooyy::::xx[bar\\foo]yy:False::::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar\\foo:: :/uri1:::::]",
    ]
    expected_gfm = """<p><img src="/uri" alt="xxbar\\fooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03xc():
    """
    Test case extra 03xc:  variation
    """

    # Arrange
    source_markdown = """[xx![bar\\foo]yy](/uri)

[bar\\foo]: /uri1"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri:::::xx![bar\\foo]yy:False::::]",
        "[text(1,2):xx:]",
        "[image(1,4):shortcut:/uri1::bar\\foo::::bar\\foo:::::]",
        "[text(1,14):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar\\foo:: :/uri1:::::]",
    ]
    expected_gfm = (
        """<p><a href="/uri">xx<img src="/uri1" alt="bar\\foo" />yy</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03a():
    """
    Test case extra 03a:  variation
    """

    # Arrange
    source_markdown = """[bar&amp;foo]: /uri

[bar&amp;foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&amp;foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar&amp;foo:::::]",
        "[text(3,2):bar\a&amp;\a\a&\a&amp;\a\afoo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar&amp;foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03aa():
    """
    Test case extra 03a:  variation
    """

    # Arrange
    source_markdown = """[xx[bar&amp;foo]yy](/uri1)

[bar&amp;foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        "[link(1,4):shortcut:/uri:::::bar&amp;foo:::::]",
        "[text(1,5):bar\a&amp;\a\a&\a&amp;\a\afoo:]",
        "[end-link::]",
        "[text(1,17):yy:]",
        "[text(1,19):]:]",
        "[text(1,20):(/uri1):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar&amp;foo:: :/uri:::::]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar&amp;foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ab():
    """
    Test case extra 03a:  variation
    """

    # Arrange
    source_markdown = """![xx[bar&amp;foo]yy](/uri1)

[bar&amp;foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:/uri1::xxbar&amp;fooyy::::xx[bar&amp;foo]yy:False::::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar&amp;foo:: :/uri:::::]",
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbar&amp;fooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ac():
    """
    Test case extra 03a:  variation
    """

    # Arrange
    source_markdown = """[xx![bar&lt;&amp;&gt;foo]yy](/uri1)

[bar&lt;&amp;&gt;foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri1:::::xx![bar&lt;&amp;&gt;foo]yy:False::::]",
        "[text(1,2):xx:]",
        "[image(1,4):shortcut:/uri::bar&lt;&amp;&gt;foo::::bar&lt;&amp;&gt;foo:::::]",
        "[text(1,26):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar&lt;&amp;&gt;foo:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri1">xx<img src="/uri" alt="bar&lt;&amp;&gt;foo" />yy</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03b():
    """
    Test case extra 03b:  variation
    """

    # Arrange
    source_markdown = """[bar&copy;foo]: /uri

[bar&copy;foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&copy;foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar&copy;foo:::::]",
        "[text(3,2):bar\a&copy;\a©\afoo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar©foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ba():
    """
    Test case extra 03b:  variation
    """

    # Arrange
    source_markdown = """[bar&copy;foo]: /uri

[xx[bar&copy;foo]yy](/uri1)"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&copy;foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):xx:]",
        "[link(3,4):shortcut:/uri:::::bar&copy;foo:::::]",
        "[text(3,5):bar\a&copy;\a©\afoo:]",
        "[end-link::]",
        "[text(3,18):yy:]",
        "[text(3,20):]:]",
        "[text(3,21):(/uri1):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar©foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03bb():
    """
    Test case extra 03b:  variation
    """

    # Arrange
    source_markdown = """[bar&copy;foo]: /uri

![xx[bar&copy;foo]yy](/uri1)"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&copy;foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[image(3,1):inline:/uri1::xxbar©fooyy::::xx[bar&copy;foo]yy:False::::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbar©fooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03bc():
    """
    Test case extra 03b:  variation
    """

    # Arrange
    source_markdown = """[bar&copy;foo]: /uri

[xx![bar&copy;foo]yy](/uri1)"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar&copy;foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):inline:/uri1:::::xx![bar&copy;foo]yy:False::::]",
        "[text(3,2):xx:]",
        "[image(3,4):shortcut:/uri::bar©foo::::bar&copy;foo:::::]",
        "[text(3,19):yy:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri1">xx<img src="/uri" alt="bar©foo" />yy</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03c():
    """
    Test case extra 03c:  variation
    """

    # Arrange
    source_markdown = """[bar` span `foo]: /uri

[bar` span `foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar` span `foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar` span `foo:::::]",
        "[text(3,2):bar:]",
        "[icode-span(3,5):span:`: : ]",
        "[text(3,13):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar<code>span</code>foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ca():
    """
    Test case extra 03ca:  variation
    """

    # Arrange
    source_markdown = """[xx[bar` span `foo]yy](/uri1)

[bar` span `foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        "[link(1,4):shortcut:/uri:::::bar` span `foo:::::]",
        "[text(1,5):bar:]",
        "[icode-span(1,8):span:`: : ]",
        "[text(1,16):foo:]",
        "[end-link::]",
        "[text(1,20):yy:]",
        "[text(1,22):]:]",
        "[text(1,23):(/uri1):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar` span `foo:: :/uri:::::]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar<code>span</code>foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03cb():
    """
    Test case extra 03cb:  variation
    """

    # Arrange
    source_markdown = """![xx[bar` span `foo]yy](/uri1)

[bar` span `foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:/uri1::xxbarspanfooyy::::xx[bar` span `foo]yy:False::::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar` span `foo:: :/uri:::::]",
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbarspanfooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03cc():
    """
    Test case extra 03cb:  variation
    """

    # Arrange
    source_markdown = """[xx![bar` span `foo]yy](/uri1)

[bar` span `foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri1:::::xx![bar` span `foo]yy:False::::]",
        "[text(1,2):xx:]",
        "[image(1,4):shortcut:/uri::barspanfoo::::bar` span `foo:::::]",
        "[text(1,21):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar` span `foo:: :/uri:::::]",
    ]
    expected_gfm = (
        """<p><a href="/uri1">xx<img src="/uri" alt="barspanfoo" />yy</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03d():
    """
    Test case extra 03c:  variation
    """

    # Arrange
    source_markdown = """[bar*span*foo]: /uri

[bar*span*foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar*span*foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar*span*foo:::::]",
        "[text(3,2):bar:]",
        "[emphasis(3,5):1:*]",
        "[text(3,6):span:]",
        "[end-emphasis(3,10)::]",
        "[text(3,11):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar<em>span</em>foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03da():
    """
    Test case extra 03c:  variation
    """

    # Arrange
    source_markdown = """[xx[bar*span*foo]yy](/uri1)

[bar*span*foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        "[link(1,4):shortcut:/uri:::::bar*span*foo:::::]",
        "[text(1,5):bar:]",
        "[emphasis(1,8):1:*]",
        "[text(1,9):span:]",
        "[end-emphasis(1,13)::]",
        "[text(1,14):foo:]",
        "[end-link::]",
        "[text(1,18):yy:]",
        "[text(1,20):]:]",
        "[text(1,21):(/uri1):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar*span*foo:: :/uri:::::]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar<em>span</em>foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03db():
    """
    Test case extra 03c:  variation
    """

    # Arrange
    source_markdown = """![xx[bar*span*foo]yy](/uri1)

[bar*span*foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:/uri1::xxbarspanfooyy::::xx[bar*span*foo]yy:False::::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar*span*foo:: :/uri:::::]",
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbarspanfooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03dc():
    """
    Test case extra 03c:  variation
    """

    # Arrange
    source_markdown = """[xx![bar*span*foo]yy](/uri1)

[bar*span*foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri1:::::xx![bar*span*foo]yy:False::::]",
        "[text(1,2):xx:]",
        "[image(1,4):shortcut:/uri::barspanfoo::::bar*span*foo:::::]",
        "[text(1,19):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar*span*foo:: :/uri:::::]",
    ]
    expected_gfm = (
        """<p><a href="/uri1">xx<img src="/uri" alt="barspanfoo" />yy</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03e():
    """
    Test case extra 03e:  variation
    """

    # Arrange
    source_markdown = """[bar<http://autolink.com>foo]: /uri

[bar<http://autolink.com>foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar<http://autolink.com>foo:: :/uri:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/uri:::::bar<http://autolink.com>foo:::::]",
        "[text(3,2):bar:]",
        "[uri-autolink(3,5):http://autolink.com]",
        "[text(3,26):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar<a href="http://autolink.com">http://autolink.com</a>foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ea():
    """
    Test case extra 03e:  variation
    """

    # Arrange
    source_markdown = """[xx[bar<http://autolink.com>foo]yy](/uri1)

[bar<http://autolink.com>foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        "[link(1,4):shortcut:/uri:::::bar<http://autolink.com>foo:::::]",
        "[text(1,5):bar:]",
        "[uri-autolink(1,8):http://autolink.com]",
        "[text(1,29):foo:]",
        "[end-link::]",
        "[text(1,33):yy:]",
        "[text(1,35):]:]",
        "[text(1,36):(/uri1):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar<http://autolink.com>foo:: :/uri:::::]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar<a href="http://autolink.com">http://autolink.com</a>foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03eb():
    """
    Test case extra 03e:  variation
    """

    # Arrange
    source_markdown = """![xx[bar<http://autolink.com>foo]yy](/uri1)

[bar<http://autolink.com>foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:/uri1::xxbarhttp://autolink.comfooyy::::xx[bar<http://autolink.com>foo]yy:False::::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar<http://autolink.com>foo:: :/uri:::::]",
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbarhttp://autolink.comfooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ec():
    """
    Test case extra 03e:  variation
    """

    # Arrange
    source_markdown = """[xx![bar<http://autolink.com>foo]yy](/uri1)

[bar<http://autolink.com>foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri1:::::xx![bar<http://autolink.com>foo]yy:False::::]",
        "[text(1,2):xx:]",
        "[image(1,4):shortcut:/uri::barhttp://autolink.comfoo::::bar<http://autolink.com>foo:::::]",
        "[text(1,34):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::bar<http://autolink.com>foo:: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri1">xx<img src="/uri" alt="barhttp://autolink.comfoo" />yy</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03f():
    """
    Test case extra 03f:  variation
    """

    # Arrange
    source_markdown = """[bar<image src="xx">foo]: /uri

[bar<image src="xx">foo]"""
    expected_tokens = [
        '[link-ref-def(1,1):True::bar<image src="xx">foo:: :/uri:::::]',
        "[BLANK(2,1):]",
        "[para(3,1):]",
        '[link(3,1):shortcut:/uri:::::bar<image src="xx">foo:::::]',
        "[text(3,2):bar:]",
        '[raw-html(3,5):image src="xx"]',
        "[text(3,21):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar<image src="xx">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03fa():
    """
    Test case extra 03f:  variation
    """

    # Arrange
    source_markdown = """[xx[bar<image src="xx">foo]yy](/uri1)

[bar<image src="xx">foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        '[link(1,4):shortcut:/uri:::::bar<image src="xx">foo:::::]',
        "[text(1,5):bar:]",
        '[raw-html(1,8):image src="xx"]',
        "[text(1,24):foo:]",
        "[end-link::]",
        "[text(1,28):yy:]",
        "[text(1,30):]:]",
        "[text(1,31):(/uri1):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar<image src="xx">foo:: :/uri:::::]',
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar<image src="xx">foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03fb():
    """
    Test case extra 03f:  variation
    """

    # Arrange
    source_markdown = """![xx[bar<image src="xx">foo]yy](/uri1)

[bar<image src="xx">foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        '[image(1,1):inline:/uri1::xxbar<image src="xx">fooyy::::xx[bar<image src="xx">foo]yy:False::::]',
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar<image src="xx">foo:: :/uri:::::]',
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbar<image src="xx">fooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03fc():
    """
    Test case extra 03f:  variation
    """

    # Arrange
    source_markdown = """[xx![bar<image src="xx">foo]yy](/uri1)

[bar<image src="xx">foo]: /uri"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/uri1:::::xx![bar<image src="xx">foo]yy:False::::]',
        "[text(1,2):xx:]",
        '[image(1,4):shortcut:/uri::bar<image src="xx">foo::::bar<image src="xx">foo:::::]',
        "[text(1,29):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::bar<image src="xx">foo:: :/uri:::::]',
    ]
    expected_gfm = """<p><a href="/uri1">xx<img src="/uri" alt="bar<image src="xx">foo" />yy</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03g():
    """
    Test case extra 03g:  variation
    """

    # Arrange
    source_markdown = """[bar
foo]: /uri

[bar\nfoo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar foo:bar\nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[link(4,1):shortcut:/uri:::::bar\nfoo:::::]",
        "[text(4,2):bar\nfoo::\n]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar
foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ga():
    """
    Test case extra 03g:  variation
    """

    # Arrange
    source_markdown = """[xx[bar
foo]yy](/uri1)

[bar
foo]: /uri
"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):xx:]",
        "[link(1,4):shortcut:/uri:::::bar\nfoo:::::]",
        "[text(1,5):bar\nfoo::\n]",
        "[end-link::]",
        "[text(2,5):yy:]",
        "[text(2,7):]:]",
        "[text(2,8):(/uri1):]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar foo:bar\nfoo: :/uri:::::]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar
foo</a>yy](/uri1)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03gb():
    """
    Test case extra 03g:  variation
    """

    # Arrange
    source_markdown = """![xx[bar
foo]yy](/uri1)

[bar
foo]: /uri
"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[image(1,1):inline:/uri1::xxbar\nfooyy::::xx[bar\nfoo]yy:False::::]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar foo:bar\nfoo: :/uri:::::]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<p><img src="/uri1" alt="xxbar
fooyy" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03gc():
    """
    Test case extra 03g:  variation
    """

    # Arrange
    source_markdown = """[xx![bar
foo]yy](/uri1)

[bar
foo]: /uri"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[link(1,1):inline:/uri1:::::xx![bar\nfoo]yy:False::::]",
        "[text(1,2):xx:]",
        "[image(1,4):shortcut:/uri::bar\nfoo::::bar\nfoo:::::]",
        "[text(2,5):yy:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar foo:bar\nfoo: :/uri:::::]",
    ]
    expected_gfm = """<p><a href="/uri1">xx<img src="/uri" alt="bar
foo" />yy</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03hx():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\
foo]: /uri

[bar\\
foo]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):bar:]",
        "[hard-break(1,5):\\:\n]",
        "[text(2,1):foo:]",
        "[text(2,4):]:]",
        "[text(2,5):: /uri:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[text(4,2):bar:]",
        "[hard-break(4,5):\\:\n]",
        "[text(5,1):foo:]",
        "[text(5,4):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[bar<br />
foo]: /uri</p>
<p>[bar<br />
foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03hxa():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[\\
foo]: /uri

[\\
foo]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[hard-break(1,2):\\:\n]",
        "[text(2,1):foo:]",
        "[text(2,4):]:]",
        "[text(2,5):: /uri:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[hard-break(4,2):\\:\n]",
        "[text(5,1):foo:]",
        "[text(5,4):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[<br />
foo]: /uri</p>
<p>[<br />
foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03hy():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[b\\ar\\
foo]: /uri

[b\\ar\\
foo]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):b\\ar:]",
        "[hard-break(1,6):\\:\n]",
        "[text(2,1):foo:]",
        "[text(2,4):]:]",
        "[text(2,5):: /uri:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[text(4,2):b\\ar:]",
        "[hard-break(4,6):\\:\n]",
        "[text(5,1):foo:]",
        "[text(5,4):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[b\\ar<br />
foo]: /uri</p>
<p>[b\\ar<br />
foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03hz():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\\\
foo]: /uri

[bar\\\\
foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar\\\\ foo:bar\\\\\nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[link(4,1):shortcut:/uri:::::bar\\\\\nfoo:::::]",
        "[text(4,2):bar\\\b\\\nfoo::\n]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar\\
foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ha():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\
foo]: /uri

[xx[bar\\
foo]yy]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):bar:]",
        "[hard-break(1,5):\\:\n]",
        "[text(2,1):foo:]",
        "[text(2,4):]:]",
        "[text(2,5):: /uri:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[text(4,2):xx:]",
        "[text(4,4):[:]",
        "[text(4,5):bar:]",
        "[hard-break(4,8):\\:\n]",
        "[text(5,1):foo:]",
        "[text(5,4):]:]",
        "[text(5,5):yy:]",
        "[text(5,7):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[bar<br />
foo]: /uri</p>
<p>[xx[bar<br />
foo]yy]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03hb():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\
foo]: /uri

![xx[bar\\
foo]yy]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):bar:]",
        "[hard-break(1,5):\\:\n]",
        "[text(2,1):foo:]",
        "[text(2,4):]:]",
        "[text(2,5):: /uri:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):![:]",
        "[text(4,3):xx:]",
        "[text(4,5):[:]",
        "[text(4,6):bar:]",
        "[hard-break(4,9):\\:\n]",
        "[text(5,1):foo:]",
        "[text(5,4):]:]",
        "[text(5,5):yy:]",
        "[text(5,7):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[bar<br />
foo]: /uri</p>
<p>![xx[bar<br />
foo]yy]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03hc():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\
foo]: /uri

[xx![bar\\
foo]yy]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):bar:]",
        "[hard-break(1,5):\\:\n]",
        "[text(2,1):foo:]",
        "[text(2,4):]:]",
        "[text(2,5):: /uri:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[text(4,2):xx:]",
        "[text(4,4):![:]",
        "[text(4,6):bar:]",
        "[hard-break(4,9):\\:\n]",
        "[text(5,1):foo:]",
        "[text(5,4):]:]",
        "[text(5,5):yy:]",
        "[text(5,7):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[bar<br />
foo]: /uri</p>
<p>[xx![bar<br />
foo]yy]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03i():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\a\a
foo]: /uri

[bar\a\a
foo]""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[link-ref-def(1,1):True::bar foo:bar  \nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[link(4,1):shortcut:/uri:::::bar  \nfoo:::::]",
        "[text(4,2):bar:]",
        "[hard-break(4,5):  :\n]",
        "[text(5,1):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">bar<br />
foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ia():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\a\a
foo]: /uri

[xx[bar\a\a
foo]yy]""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[link-ref-def(1,1):True::bar foo:bar  \nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[text(4,2):xx:]",
        "[link(4,4):shortcut:/uri:::::bar  \nfoo:::::]",
        "[text(4,5):bar:]",
        "[hard-break(4,8):  :\n]",
        "[text(5,1):foo:]",
        "[end-link::]",
        "[text(5,5):yy:]",
        "[text(5,7):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[xx<a href="/uri">bar<br />
foo</a>yy]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ib():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\a\a
foo]: /uri

![xx[bar\a\a
foo]yy]""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[link-ref-def(1,1):True::bar foo:bar  \nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):![:]",
        "[text(4,3):xx:]",
        "[link(4,5):shortcut:/uri:::::bar  \nfoo:::::]",
        "[text(4,6):bar:]",
        "[hard-break(4,9):  :\n]",
        "[text(5,1):foo:]",
        "[end-link::]",
        "[text(5,5):yy:]",
        "[text(5,7):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>![xx<a href="/uri">bar<br />
foo</a>yy]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ic():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\a\a
foo]: /uri

[xx![bar\a\a
foo]yy]""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[link-ref-def(1,1):True::bar foo:bar  \nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n]",
        "[text(4,1):[:]",
        "[text(4,2):xx:]",
        "[image(4,4):shortcut:/uri::bar\nfoo::::bar  \nfoo:::::]",
        "[text(5,5):yy:]",
        "[text(5,7):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[xx<img src="/uri" alt="bar
foo" />yy]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03jx():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\\\
foo]: /uri

abc
[bar\\\\
foo]
abc"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar\\\\ foo:bar\\\\\nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n\n\n]",
        "[text(4,1):abc\n::\n]",
        "[link(5,1):shortcut:/uri:::::bar\\\\\nfoo:::::]",
        "[text(5,2):bar\\\b\\\nfoo::\n]",
        "[end-link::]",
        "[text(6,5):\nabc::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>abc
<a href="/uri">bar\\
foo</a>
abc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_reference_links_extra_03ja():
    """
    Test case extra 03h:  variation
    """

    # Arrange
    source_markdown = """[bar\\\\
foo]: /uri

abc
[bar\\\\
foo][]
abc"""
    expected_tokens = [
        "[link-ref-def(1,1):True::bar\\\\ foo:bar\\\\\nfoo: :/uri:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):\n\n\n]",
        "[text(4,1):abc\n::\n]",
        "[link(5,1):collapsed:/uri:::::bar\\\\\nfoo:::::]",
        "[text(5,2):bar\\\b\\\nfoo::\n]",
        "[end-link::]",
        "[text(6,7):\nabc::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>abc
<a href="/uri">bar\\
foo</a>
abc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide classes to deal with plugins.
"""
import argparse
import inspect
import logging
import os
import re
import sys
from abc import ABC, abstractmethod

from columnar import columnar

from pymarkdown.application_properties import ApplicationPropertiesFacade
from pymarkdown.extensions.pragma_token import PragmaToken
from pymarkdown.parser_helper import ParserHelper

LOGGER = logging.getLogger(__name__)

# pylint: disable=too-many-lines


# pylint: disable=too-few-public-methods
class ScanContext:
    """
    Class to provide context when reporting any errors.
    """

    def __init__(self, owning_manager, scan_file):
        self.owning_manager, self.scan_file, self.line_number = (
            owning_manager,
            scan_file,
            0,
        )


# pylint: enable=too-few-public-methods


class BadPluginError(Exception):
    """
    Class to allow for a critical error within a plugin to be encapsulated
    and reported.
    """

    # pylint: disable=too-many-arguments, too-many-branches
    def __init__(
        self,
        plugin_id=None,
        plugin_action=None,
        file_name=None,
        class_name=None,
        field_name=None,
        is_constructor=False,
        is_empty=False,
        formatted_message=None,
        line_number=0,
        column_number=0,
        actual_line=None,
        actual_token=None,
        cause=None,
    ):

        if not formatted_message:
            if file_name:
                if class_name:
                    if is_constructor:
                        formatted_message = f"Plugin file named '{file_name}' threw an exception in the constructor for the class '{class_name}'."
                    else:
                        formatted_message = f"Plugin file named '{file_name}' does not contain a class named '{class_name}'."
                else:
                    formatted_message = (
                        f"Plugin file named '{file_name}' cannot be loaded."
                    )
            elif class_name:
                if field_name:
                    if is_empty:
                        formatted_message = f"Plugin class '{class_name}' returned an empty value for field name '{field_name}'."
                    else:
                        formatted_message = f"Plugin class '{class_name}' returned an improperly typed value for field name '{field_name}'."
                else:
                    formatted_message = f"Plugin class '{class_name}' had a critical failure loading the plugin details."
            else:
                if cause and isinstance(cause, ValueError):
                    formatted_message = str(cause)
                else:
                    formatted_message = f"Plugin id '{plugin_id.upper()}' had a critical failure during the '{str(plugin_action)}' action."
            if line_number:
                position_message = (
                    f"({line_number},{column_number})"
                    if column_number
                    else f"(Line {line_number})"
                )
                formatted_message = f"{position_message}: {formatted_message}"
            if actual_line:
                formatted_message = f"{formatted_message}\nActual Line: {actual_line}"
            if actual_token:
                formatted_message = f"{formatted_message}\nActual Token: {ParserHelper.make_value_visible(actual_token)}"
        super().__init__(formatted_message)

    # pylint: enable=too-many-arguments, too-many-branches


class Plugin(ABC):
    """
    Class to provide structure to scan through a file.
    Based off of concepts from https://github.com/hiddenillusion/example-code/commit/3e2daada652fe9b487574c784e0924bd5fcfe667
    """

    def __init__(self):
        (
            self.__plugin_specific_facade,
            self.__is_next_token_implemented_in_plugin,
            self.__is_next_line_implemented_in_plugin,
            self.__is_starting_new_file_implemented_in_plugin,
            self.__is_completed_file_implemented_in_plugin,
        ) = (None, True, True, True, True)

    @abstractmethod
    def get_details(self):
        """
        Get the details for the plugin.
        """

    @property
    def plugin_configuration(self):
        """
        Get the configuration facade associated with this plugin.
        """
        return self.__plugin_specific_facade

    def set_configuration_map(self, plugin_specific_facade):
        """
        Set the configuration map with values for the plugin.
        """
        self.__plugin_specific_facade = plugin_specific_facade

        self.__is_next_token_implemented_in_plugin = (
            "next_token" in self.__class__.__dict__.keys()
        )
        self.__is_next_line_implemented_in_plugin = (
            "next_line" in self.__class__.__dict__.keys()
        )
        self.__is_starting_new_file_implemented_in_plugin = (
            "starting_new_file" in self.__class__.__dict__.keys()
        )
        self.__is_completed_file_implemented_in_plugin = (
            "completed_file" in self.__class__.__dict__.keys()
        )

    @property
    def is_starting_new_file_implemented_in_plugin(self):
        """
        Return whether the starting_new_file function is implemented in the plugin.
        """
        return self.__is_starting_new_file_implemented_in_plugin

    @property
    def is_next_line_implemented_in_plugin(self):
        """
        Return whether the next_line function is implemented in the plugin.
        """
        return self.__is_next_line_implemented_in_plugin

    @property
    def is_next_token_implemented_in_plugin(self):
        """
        Return whether the next_token function is implemented in the plugin.
        """
        return self.__is_next_token_implemented_in_plugin

    @property
    def is_completed_file_implemented_in_plugin(self):
        """
        Return whether the completed_file function is implemented in the plugin.
        """
        return self.__is_completed_file_implemented_in_plugin

    def report_next_line_error(self, context, column_number, line_number_delta=0):
        """
        Report an error with the current line being processed.
        """
        context.owning_manager.log_scan_failure(
            context.scan_file,
            context.line_number + line_number_delta,
            column_number,
            self.get_details().plugin_id,
            self.get_details().plugin_name,
            self.get_details().plugin_description,
        )

    # pylint: disable=too-many-arguments
    def report_next_token_error(
        self,
        context,
        token,
        extra_error_information=None,
        line_number_delta=0,
        column_number_delta=0,
        use_original_position=False,
    ):
        """
        Report an error with the current token being processed.
        """
        context.owning_manager.log_scan_failure(
            context.scan_file,
            (token.original_line_number if use_original_position else token.line_number)
            + line_number_delta,
            (
                token.original_column_number
                if use_original_position
                else token.column_number
            )
            + column_number_delta
            if column_number_delta >= 0
            else -column_number_delta,
            self.get_details().plugin_id,
            self.get_details().plugin_name,
            self.get_details().plugin_description,
            extra_error_information=extra_error_information,
        )

    # pylint: enable=too-many-arguments

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """

    def completed_file(self, context):
        """
        Event that the file being currently scanned is now completed.
        """

    def next_line(self, context, line):
        """
        Event that a new line is being processed.
        """

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """


# pylint: disable=too-few-public-methods
class PluginDetails:
    """
    Class to provide details about a plugin, supplied by the plugin.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        plugin_id,
        plugin_name,
        plugin_description,
        plugin_enabled_by_default,
        plugin_version,
        plugin_interface_version,
    ):
        (
            self.plugin_id,
            self.plugin_name,
            self.plugin_description,
            self.plugin_enabled_by_default,
            self.plugin_version,
            self.plugin_interface_version,
        ) = (
            plugin_id,
            plugin_name,
            plugin_description,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
        )

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class FoundPlugin:
    """
    Encapsulation of a plugin that was discovered.  While similar to the PluginDetails
    class, this is meant for an internal representation of the plugin, and not the
    external information provided.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        plugin_id,
        plugin_name,
        plugin_description,
        plugin_instance,
        plugin_enabled_by_default,
        plugin_version,
        plugin_interface_version,
        instance_file_name,
    ):
        """
        Initializes a new instance of the FoundPlugin class.
        """
        (
            self.__plugin_id,
            self.__plugin_names,
            self.__plugin_description,
            self.__plugin_instance,
            self.__plugin_enabled_by_default,
            self.__plugin_version,
            self.__plugin_interface_version,
            self.__plugin_file_name,
        ) = (
            plugin_id.strip().lower(),
            [],
            plugin_description,
            plugin_instance,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            instance_file_name,
        )
        for next_name in plugin_name.lower().split(","):
            next_name = next_name.strip()
            if next_name:
                self.__plugin_names.append(next_name)

    # pylint: enable=too-many-arguments

    @property
    def plugin_id(self):
        """
        Gets the id associated with the plugin.
        """
        return self.__plugin_id

    @property
    def plugin_names(self):
        """
        Gets the names associated with the plugin.
        """
        return self.__plugin_names

    @property
    def plugin_identifiers(self):
        """
        Gets the identifiers (id+names) for the plugin.
        """
        plugin_keys = [self.plugin_id]
        plugin_keys.extend(self.plugin_names)
        return plugin_keys

    @property
    def plugin_description(self):
        """
        Gets the description of the plugin.
        """
        return self.__plugin_description

    @property
    def plugin_instance(self):
        """
        Gets the actual instance of the plugin.
        """
        return self.__plugin_instance

    @property
    def plugin_file_name(self):
        """
        Gets the filename where the plugin's class is stored.
        """
        return self.__plugin_file_name

    @property
    def plugin_version(self):
        """
        Gets the version of the plugin.
        """
        return self.__plugin_version

    @property
    def plugin_enabled_by_default(self):
        """
        Gets a value indicating whether the plugin is enabled by default.
        """
        return self.__plugin_enabled_by_default


# pylint: enable=too-few-public-methods,too-many-instance-attributes


# pylint: disable=too-many-instance-attributes
class PluginManager:
    """
    Manager object to take care of load and accessing plugin modules.
    """

    __plugin_prefix = "plugins"
    __root_subparser_name = "pm_subcommand"
    __argparse_subparser = None
    __id_regex = re.compile("^[a-z]{2,3}[0-9]{3,3}$")
    __name_regex = re.compile("^[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9]$")
    __filter_regex = re.compile("^[a-zA-Z0-9-]+$")
    __version_regex = re.compile("^(0|[1-9]\\d*)\\.(0|[1-9]\\d*)\\.(0|[1-9]\\d*)$")

    def __init__(self):
        (
            self.__registered_plugins,
            self.__enabled_plugins,
            self.__enabled_plugins_for_starting_new_file,
            self.__enabled_plugins_for_next_token,
            self.__enabled_plugins_for_next_line,
            self.__enabled_plugins_for_completed_file,
            self.__loaded_classes,
            self.number_of_scan_failures,
            self.number_of_pragma_failures,
            self.__show_stack_trace,
            self.__document_pragmas,
            self.__all_ids,
        ) = (None, None, None, None, None, None, None, None, None, False, None, None)

    # pylint: disable=too-many-arguments
    def initialize(
        self,
        directory_to_search,
        additional_paths,
        enable_rules_from_command_line,
        disable_rules_from_command_line,
        properties,
        show_stack_trace,
    ):
        """
        Initializes the manager by scanning for plugins, loading them, and registering them.
        """
        self.__show_stack_trace = show_stack_trace
        (
            self.number_of_scan_failures,
            self.number_of_pragma_failures,
            self.__loaded_classes,
        ) = (0, 0, [])

        plugin_files = self.__find_eligible_plugins_in_directory(directory_to_search)
        self.__load_plugins(directory_to_search, plugin_files)

        if additional_paths:
            for next_additional_plugin in additional_paths:
                if not os.path.exists(next_additional_plugin):
                    formatted_message = (
                        f"Plugin path '{next_additional_plugin}' does not exist."
                    )
                    raise BadPluginError(formatted_message=formatted_message)
                if os.path.isdir(next_additional_plugin):
                    plugin_files = self.__find_eligible_plugins_in_directory(
                        next_additional_plugin
                    )
                    self.__load_plugins(next_additional_plugin, plugin_files)
                else:
                    self.__load_plugins(
                        os.path.dirname(next_additional_plugin),
                        [os.path.basename(next_additional_plugin)],
                    )

        self.__register_plugins(
            enable_rules_from_command_line, disable_rules_from_command_line, properties
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def argparse_subparser_name():
        """
        Gets the name of the subparser used to handle these plugins.
        """
        return "plugins"

    @staticmethod
    def __list_filter_type(argument):
        test_argument = argument.replace("*", "").replace("?", "")
        if PluginManager.__filter_regex.match(test_argument):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid pattern for an id or a name."
        )

    @staticmethod
    def __info_filter_type(argument):
        if PluginManager.__id_regex.match(argument) or PluginManager.__name_regex.match(
            argument
        ):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid id or name."
        )

    @staticmethod
    def add_argparse_subparser(subparsers):
        """
        Populate the argparse tree to allow for plugin support.
        """

        new_sub_parser = subparsers.add_parser(
            PluginManager.argparse_subparser_name(), help="plugin commands"
        )
        PluginManager.__argparse_subparser = new_sub_parser
        plugin_subparsers = new_sub_parser.add_subparsers(
            dest=PluginManager.__root_subparser_name
        )

        sub_sub_parser = plugin_subparsers.add_parser(
            "list", help="list the available plugins"
        )
        sub_sub_parser.add_argument(
            "--all",
            dest="show_all",
            action="store_true",
            default=False,
            help="show all loaded plugins (default is False)",
        )
        sub_sub_parser.add_argument(
            dest="list_filter",
            default=None,
            help="filter",
            nargs="?",
            type=PluginManager.__list_filter_type,
        )
        sub_sub_parser = plugin_subparsers.add_parser(
            "info", help="information of specific plugins"
        )
        sub_sub_parser.add_argument(
            dest="info_filter",
            default=None,
            type=PluginManager.__info_filter_type,
            help="an id",
        )

    # pylint: disable=too-many-locals
    def __handle_argparse_subparser_list(self, args):
        list_re = None
        if args.list_filter:
            list_re = re.compile(
                "^" + args.list_filter.replace("*", ".*").replace("?", ".") + "$"
            )

        show_rows = []
        for next_plugin_id in self.all_plugin_ids:
            if next_plugin_id.startswith("md9"):
                continue
            next_plugin_list = []
            for next_plugin in self.__registered_plugins:
                if next_plugin.plugin_id == next_plugin_id:
                    next_plugin_list.append(next_plugin)
            assert len(next_plugin_list) == 1
            next_plugin = next_plugin_list[0]
            if next_plugin.plugin_version == "0.0.0" and not args.show_all:
                continue
            does_match = True
            if list_re:
                does_match = list_re.match(next_plugin_id)
                if not does_match:
                    for next_name in next_plugin.plugin_names:
                        does_match = list_re.match(next_name)
                        if does_match:
                            break
            if does_match:
                is_enabled_now = next_plugin in self.__enabled_plugins
                display_row = [
                    next_plugin_id,
                    ", ".join(next_plugin.plugin_names),
                    str(next_plugin.plugin_enabled_by_default),
                    str(is_enabled_now),
                    next_plugin.plugin_version,
                ]
                show_rows.append(display_row)

        headers = ["id", "names", "enabled (default)", "enabled (current)", "version"]
        table = columnar(show_rows, headers, no_borders=True)
        split_rows = table.split("\n")
        new_rows = []
        for next_row in split_rows:
            new_rows.append(next_row.rstrip())
        print("\n".join(new_rows))

    # pylint: enable=too-many-locals

    def __handle_argparse_subparser_info(self, args):
        found_plugin = list(
            filter(
                lambda x: args.info_filter in x.plugin_identifiers,
                self.__registered_plugins,
            )
        )
        if not found_plugin:
            print(
                f"Unable to find a plugin with an id or name of '{args.info_filter}'."
            )
            return 1

        found_plugin = found_plugin[0]
        print("Id:" + found_plugin.plugin_id)
        print("Name(s):" + ",".join(found_plugin.plugin_names))
        print("Description:" + found_plugin.plugin_description)
        return 0

    def handle_argparse_subparser(self, args):
        """
        Handle the parsing for this subparser.
        """
        subparser_value = getattr(args, PluginManager.__root_subparser_name)
        return_code = 0
        if subparser_value == "list":
            self.__handle_argparse_subparser_list(args)
        elif subparser_value == "info":
            return_code = self.__handle_argparse_subparser_info(args)
        else:
            PluginManager.__argparse_subparser.print_help()
            sys.exit(2)
        return return_code

    # pylint: disable=too-many-arguments
    def log_scan_failure(
        self,
        scan_file,
        line_number,
        column_number,
        rule_id,
        rule_name,
        rule_description,
        extra_error_information=None,
    ):
        """
        Log the scan failure in the appropriate format.
        """

        if self.__document_pragmas and line_number in self.__document_pragmas:
            id_set = self.__document_pragmas[line_number]
            rule_id = rule_id.lower()
            if rule_id in id_set:
                return

        extra_info = f" [{extra_error_information}]" if extra_error_information else ""

        print(
            "{0}:{1}:{2}: {3}: {4}{5} ({6})".format(
                scan_file,
                line_number,
                column_number,
                rule_id.upper(),
                rule_description,
                extra_info,
                rule_name,
            )
        )
        self.number_of_scan_failures += 1

    # pylint: enable=too-many-arguments

    def log_pragma_failure(self, scan_file, line_number, pragma_error):
        """
        Log the pragma failure in the appropriate format.
        """

        print("{0}:{1}:1: INLINE: {2}".format(scan_file, line_number, pragma_error))
        self.number_of_pragma_failures += 1

    # pylint: disable=too-many-locals
    def compile_pragmas(self, scan_file, pragma_token):
        """
        Go through the list of extracted pragmas and compile them.
        """

        for next_line_number in pragma_token.pragma_lines.keys():
            if next_line_number > 0:
                prefix_length = len(PragmaToken.pragma_prefix)
                actual_line_number = next_line_number
            else:
                prefix_length = len(PragmaToken.pragma_alternate_prefix)
                actual_line_number = -next_line_number

            line_after_prefix = pragma_token.pragma_lines[next_line_number][
                prefix_length:
            ].rstrip()
            after_whitespace_index, _ = ParserHelper.extract_whitespace(
                line_after_prefix, 0
            )
            command_data = line_after_prefix[
                after_whitespace_index
                + len(PragmaToken.pragma_title) : -len(PragmaToken.pragma_suffix)
            ]
            after_command_index, command = ParserHelper.extract_until_whitespace(
                command_data, 0
            )
            command = command.lower()
            if not command:
                self.log_pragma_failure(
                    scan_file,
                    actual_line_number,
                    "Inline configuration specified without command.",
                )
            elif command == "disable-next-line":
                ids_to_disable = command_data[after_command_index:].split(",")
                processed_ids = set()
                for next_id in ids_to_disable:
                    next_id = next_id.strip().lower()
                    if not next_id:
                        self.log_pragma_failure(
                            scan_file,
                            actual_line_number,
                            f"Inline configuration command '{command}' specified a plugin with a blank id.",
                        )
                    elif next_id in self.__all_ids:
                        normalized_id = self.__all_ids[next_id].plugin_id
                        processed_ids.add(normalized_id)
                    else:
                        self.log_pragma_failure(
                            scan_file,
                            actual_line_number,
                            f"Inline configuration command '{command}' unable to find a plugin with the id '{next_id}'.",
                        )

                if processed_ids:
                    self.__document_pragmas[actual_line_number + 1] = processed_ids
            else:
                self.log_pragma_failure(
                    scan_file,
                    actual_line_number,
                    f"Inline configuration command '{command}' not understood.",
                )

    # pylint: enable=too-many-locals

    @classmethod
    def __find_eligible_plugins_in_directory(cls, directory_to_search):
        """
        Given a directory to search, scan for eligible modules to load later.
        """

        plugin_files = [
            x
            for x in os.listdir(directory_to_search)
            if x.endswith(".py") and x[0:-3] != "__init__"
        ]
        return plugin_files

    @classmethod
    def __snake_to_camel(cls, word):

        return "".join(x.capitalize() or "_" for x in word.split("_"))

    def __attempt_to_load_plugin(
        self, next_plugin_module, plugin_class_name, next_plugin_file
    ):
        """
        Attempt to cleanly load the specified plugin.
        """
        try:
            mod = __import__(next_plugin_module)
        except Exception as this_exception:
            raise BadPluginError(file_name=next_plugin_file) from this_exception

        if not hasattr(mod, plugin_class_name):
            raise BadPluginError(
                file_name=next_plugin_file, class_name=plugin_class_name
            ) from None
        my_class = getattr(mod, plugin_class_name)

        try:
            plugin_class_instance = my_class()
        except Exception as this_exception:
            raise BadPluginError(
                file_name=next_plugin_file,
                class_name=plugin_class_name,
                is_constructor=True,
            ) from this_exception
        self.__loaded_classes.append((plugin_class_instance, next_plugin_file))

    def __load_plugins(self, directory_to_search, plugin_files):
        """
        Given an array of discovered modules, load them into the global namespace.
        """

        if os.path.abspath(directory_to_search) not in sys.path:
            sys.path.insert(0, os.path.abspath(directory_to_search))

        for next_plugin_file in plugin_files:
            next_plugin_module = next_plugin_file[0:-3]
            plugin_class_name = self.__snake_to_camel(next_plugin_module)
            self.__attempt_to_load_plugin(
                next_plugin_module, plugin_class_name, next_plugin_file
            )

    def __determine_if_plugin_enabled(
        self,
        plugin_object,
        command_line_enabled_rules,
        command_line_disabled_rules,
        properties,
    ):
        """
        Given the enable and disable rules values, evaluate the enabled or disabled
        state of the plugin in proper order.
        """

        new_value = None
        LOGGER.debug(
            "Plugin '%s', identifiers: %s",
            plugin_object.plugin_id,
            plugin_object.plugin_identifiers,
        )

        if command_line_disabled_rules:
            LOGGER.debug(
                "Disabled on command line: %s", str(command_line_disabled_rules)
            )
            for next_identifier in plugin_object.plugin_identifiers:
                if next_identifier in command_line_disabled_rules:
                    new_value = False
                    LOGGER.debug("Plugin is disabled from command line.")
                    break
        if new_value is None and command_line_enabled_rules:
            LOGGER.debug("Enabled on command line: %s", str(command_line_enabled_rules))
            for next_identifier in plugin_object.plugin_identifiers:
                if next_identifier in command_line_enabled_rules:
                    new_value = True
                    LOGGER.debug("Plugin is enabled from command line.")
                    break
        if new_value is None:
            plugin_specific_facade = self.__find_configuration_for_plugin(
                plugin_object, properties
            )
            if plugin_specific_facade:
                LOGGER.debug(
                    "Plugins specific configuration found, searching for key 'enabled'."
                )
                new_value = plugin_specific_facade.get_boolean_property(
                    "enabled", default_value=None
                )
                if new_value:
                    LOGGER.debug(
                        "Plugin specific key 'enabled' found, value is '%s'.",
                        str(new_value),
                    )
        if new_value is None:
            LOGGER.debug(
                "No other enable state found, setting to default of '%s'.",
                str(plugin_object.plugin_enabled_by_default),
            )

        return (
            plugin_object.plugin_enabled_by_default if new_value is None else new_value
        )

    # pylint: enable=too-many-arguments

    @classmethod
    def __verify_string_field(cls, plugin_instance, field_name, field_value):
        """
        Verify that the detail field is a valid string.
        """

        if not isinstance(field_value, str):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )
        if not field_value:
            raise BadPluginError(
                class_name=type(plugin_instance).__name__,
                field_name=field_name,
                is_empty=True,
            )

    @classmethod
    def __verify_boolean_field(cls, plugin_instance, field_name, field_value):
        """
        Verify that the detail field is a valid boolean.
        """

        if not isinstance(field_value, bool):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )

    @classmethod
    def __verify_integer_field(cls, plugin_instance, field_name, field_value):
        """
        Verify that the detail field is a valid integer.
        """

        if not isinstance(field_value, int):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )

    def __get_plugin_details(self, plugin_instance, instance_file_name):
        """
        Query the plugin for details and verify that they are reasonable.
        """

        try:
            instance_details = plugin_instance.get_details()
            (
                plugin_id,
                plugin_name,
                plugin_description,
                plugin_enabled_by_default,
                plugin_version,
                plugin_interface_version,
            ) = (
                instance_details.plugin_id,
                instance_details.plugin_name,
                instance_details.plugin_description,
                instance_details.plugin_enabled_by_default,
                instance_details.plugin_version,
                instance_details.plugin_interface_version,
            )
        except Exception as this_exception:
            raise BadPluginError(
                class_name=type(plugin_instance).__name__,
            ) from this_exception

        self.__verify_string_field(plugin_instance, "plugin_id", plugin_id)
        self.__verify_string_field(plugin_instance, "plugin_name", plugin_name)
        self.__verify_string_field(
            plugin_instance, "plugin_description", plugin_description
        )
        self.__verify_boolean_field(
            plugin_instance, "plugin_enabled_by_default", plugin_enabled_by_default
        )
        self.__verify_string_field(plugin_instance, "plugin_version", plugin_version)
        self.__verify_integer_field(
            plugin_instance, "plugin_interface_version", plugin_interface_version
        )
        if plugin_interface_version != 1:
            raise BadPluginError(
                formatted_message=f"Plugin '{instance_file_name}' with an interface version ('{plugin_interface_version}') that is not '1'."
            )

        plugin_object = FoundPlugin(
            plugin_id,
            plugin_name,
            plugin_description,
            plugin_instance,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            instance_file_name,
        )
        return plugin_object

    # pylint: disable=too-many-arguments
    def __register_individual_plugin(
        self,
        plugin_instance,
        instance_file_name,
        command_line_enabled_rules,
        command_line_disabled_rules,
        properties,
    ):
        """
        Register an individual plugin for use.
        """

        plugin_object = self.__get_plugin_details(plugin_instance, instance_file_name)

        next_key = plugin_object.plugin_id
        if not PluginManager.__id_regex.match(next_key):
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with id '{next_key}' as id is not a valid id in the form 'aannn' or 'aaannn'."
            )

        if next_key in self.__all_ids:
            found_plugin = self.__all_ids[next_key]
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with id '{next_key}' as plugin '{found_plugin.plugin_file_name}' is already registered with that id."
            )
        self.__all_ids[next_key] = plugin_object

        for next_key in plugin_object.plugin_names:
            if not PluginManager.__name_regex.match(next_key):
                raise ValueError(
                    f"Unable to register plugin '{instance_file_name}' with name '{next_key}' as name is not a valid name in the form 'an-an'."
                )
            if next_key in self.__all_ids:
                found_plugin = self.__all_ids[next_key]
                raise ValueError(
                    f"Unable to register plugin '{instance_file_name}' with name '{next_key}' as plugin '{found_plugin.plugin_file_name}' is already registered with that name."
                )
            self.__all_ids[next_key] = plugin_object
        if not plugin_object.plugin_description.strip():
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with a description string that is blank."
            )
        if not PluginManager.__version_regex.match(plugin_object.plugin_version):
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with a version string that is not a valid semantic version."
            )

        self.__registered_plugins.append(plugin_object)
        if self.__determine_if_plugin_enabled(
            plugin_object,
            command_line_enabled_rules,
            command_line_disabled_rules,
            properties,
        ):
            self.__enabled_plugins.append(plugin_object)

    # pylint: enable=too-many-arguments

    def __register_plugins(
        self,
        enable_rules_from_command_line,
        disable_rules_from_command_line,
        properties,
    ):
        """
        Scan the global namespace for all subclasses of the 'Plugin' class to use as
        plugins.
        """

        (
            command_line_enabled_rules,
            command_line_disabled_rules,
            self.__registered_plugins,
            self.__enabled_plugins,
            self.__all_ids,
        ) = (set(), set(), [], [], {})
        if enable_rules_from_command_line:
            for next_rule_identifier in enable_rules_from_command_line.lower().split(
                ","
            ):
                command_line_enabled_rules.add(next_rule_identifier)
        if disable_rules_from_command_line:
            for next_rule_identifier in disable_rules_from_command_line.lower().split(
                ","
            ):
                command_line_disabled_rules.add(next_rule_identifier)

        for plugin_instance, instance_file_name in self.__loaded_classes:
            self.__register_individual_plugin(
                plugin_instance,
                instance_file_name,
                command_line_enabled_rules,
                command_line_disabled_rules,
                properties,
            )

    @property
    def all_plugin_ids(self):
        """
        Get a list of all plugins by their id.
        """

        id_list = []
        for next_plugin in self.__registered_plugins:
            id_list.append(next_plugin.plugin_id)
        return id_list

    @classmethod
    def __find_configuration_for_plugin(
        cls, next_plugin, properties, always_return_facade=False
    ):
        plugin_specific_facade = None
        first_facade = None
        for next_key_name in next_plugin.plugin_identifiers:
            plugin_section_title = f"{PluginManager.__plugin_prefix}{properties.separator}{next_key_name}{properties.separator}"
            section_facade_candidate = ApplicationPropertiesFacade(
                properties, plugin_section_title
            )
            if not first_facade:
                first_facade = section_facade_candidate
            if section_facade_candidate.property_names:
                plugin_specific_facade = section_facade_candidate
                break

        if always_return_facade and not plugin_specific_facade:
            plugin_specific_facade = first_facade
        return plugin_specific_facade

    def apply_configuration(self, properties):
        """
        Apply any supplied configuration to each of the enabled plugins.
        """

        (
            self.__enabled_plugins_for_starting_new_file,
            self.__enabled_plugins_for_next_token,
            self.__enabled_plugins_for_next_line,
            self.__enabled_plugins_for_completed_file,
        ) = ([], [], [], [])

        for next_plugin in self.__enabled_plugins:
            try:
                plugin_specific_facade = self.__find_configuration_for_plugin(
                    next_plugin, properties, always_return_facade=True
                )

                next_plugin.plugin_instance.set_configuration_map(
                    plugin_specific_facade
                )
                next_plugin.plugin_instance.initialize_from_config()
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

            if next_plugin.plugin_instance.is_next_token_implemented_in_plugin:
                self.__enabled_plugins_for_next_token.append(next_plugin)
            if next_plugin.plugin_instance.is_next_line_implemented_in_plugin:
                self.__enabled_plugins_for_next_line.append(next_plugin)
            if next_plugin.plugin_instance.is_completed_file_implemented_in_plugin:
                self.__enabled_plugins_for_completed_file.append(next_plugin)
            if next_plugin.plugin_instance.is_starting_new_file_implemented_in_plugin:
                self.__enabled_plugins_for_starting_new_file.append(next_plugin)

    def starting_new_file(self, file_being_started):
        """
        Inform any listeners that a new current file has been started.
        """
        self.__document_pragmas = {}
        for next_plugin in self.__enabled_plugins_for_starting_new_file:
            try:
                next_plugin.plugin_instance.starting_new_file()
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

        return ScanContext(self, file_being_started)

    def completed_file(self, context, line_number):
        """
        Inform any listeners that the current file has been completed.
        """
        context.line_number = line_number
        for next_plugin in self.__enabled_plugins_for_completed_file:
            try:
                next_plugin.plugin_instance.completed_file(context)
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

    def next_line(self, context, line_number, line):
        """
        Inform any listeners that a new line has been loaded.
        """
        context.line_number = line_number
        for next_plugin in self.__enabled_plugins_for_next_line:
            try:
                next_plugin.plugin_instance.next_line(context, line)
            except Exception as this_exception:
                actual_line = line if self.__show_stack_trace else None

                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    line_number=line_number,
                    actual_line=actual_line,
                    cause=this_exception,
                ) from this_exception

    def next_token(self, context, token):
        """
        Inform any listeners of a new token that has been processed.
        """
        for next_plugin in self.__enabled_plugins_for_next_token:
            try:
                next_plugin.plugin_instance.next_token(context, token)
            except Exception as this_exception:
                actual_token = token if self.__show_stack_trace else None

                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    line_number=token.line_number,
                    column_number=token.column_number,
                    actual_token=actual_token,
                    cause=this_exception,
                ) from this_exception


# pylint: enable=too-many-instance-attributes
